@@ -1,5 +1,7 @@
+from copy import copy
 import json
 import os
+import time
 from typing import Counter
 import xlwings as xw
 import shutil
@@ -7,14 +9,37 @@ from tqdm import tqdm
 import datetime
 from interval import Interval
 import pandas as pd
+import locale
+from PIL import ImageGrab
+import win32clipboard as wc
+locale.setlocale(locale.LC_CTYPE, 'chinese')
+os.system("echo off | clip")
+
+# 回调
+class BzttExcelEvent:
+
+    logChanged = []
+    def __init__(self) -> None:
+        print("BzttExcelEvent")
+        self.logChanged = []
+
+    # 日志回调
+    def raiseLogEvent(self,*args):
+        for fun in self.logChanged:
+            fun(*args)
+
 
 class BzttExcl:  #工单处理
     log_lv = 3
-
+    excel_data = {} # 供应外部使用
     app=""
     wb = {}
+    myBzttExcelEvent = None
+
     # def __init__(self, system_info, oltInfo, soft_ware_info):
     def  __init__(self,system_info={}):
+        self.myBzttExcelEvent = BzttExcelEvent()
+
         log_lv_data = {
             "ALL": 0,
             "TRACE": 1,
@@ -72,6 +97,16 @@ class BzttExcl:  #工单处理
                 info.setdefault("end", "\n")
 
                 print("\033[1;%sm%s\033[0m" % (info["color"], info["text"]), end=info["end"])
+                time1_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+                return_data = {
+                    # "step_type":info["step_type"],# 当前步骤是否是下一个  |next,""
+                    "end":info["end"],          # 结束标志
+                    "log_lv":log_lv_num,        # 日志级别
+                    "content":info["text"], # 日志内容
+                    "do_time":time1_str      #执行时间
+                }
+
+                self.myBzttExcelEvent.raiseLogEvent(return_data)
 
 
     # excel 判断表格内容（ifs）
@@ -121,17 +156,88 @@ class BzttExcl:  #工单处理
         ]
         self.print_log(print_data, "INFO")
         sourceRange.AutoFill(fillRange, 0)
+    def WriteValue(self,jsonData):
+        # print("WriteValue",jsonData)
+        for value_info in jsonData["value"]["data"]:
+            sourceInfo = value_info["sourceInfo"]
+            sht = self.wb[sourceInfo["title"]].sheets(sourceInfo["sheet"])
+            sht.range(sourceInfo["range"]["startLetter"] + str(sourceInfo["range"]["startNum"] )).value = value_info["ValueData"]
+            pass
+
+        #.value
+    # 填充公式
+    def WriteGsData(self,jsonData):
+        print("jsonData",jsonData)
+        jsonData_value_sourceInfo = jsonData["value"]["sourceInfo"]
+        jsonData_value_gsInfo = jsonData["value"]["gsInfo"]
+        sht = self.wb[jsonData_value_sourceInfo["title"]].sheets(jsonData_value_sourceInfo["sheet"])
+        info = sht.used_range
+        nrows = info.last_cell.row
+        ncolumns = info.last_cell.column
+        # print("nrows",nrows,"ncolumns",ncolumns)
+
+        # print("正在计算请稍后...")
+        startNum = int(jsonData_value_sourceInfo["range"]["startNum"])
+        startLetter = jsonData_value_sourceInfo["range"]["startLetter"]
+
+        startNum_old = copy(startNum)
+        i = 0
+        # 填写内容
+        for gsInfo in jsonData_value_gsInfo["data"]:
+            print_data = [
+                {"text": "计算公式：" + gsInfo, "color": "34"},
+
+            ]
+            self.print_log(print_data, "DEBUG")
+            sht.range(startLetter + str(startNum)).value = gsInfo
+            startNum = startNum + 1
+            i = i + 1
+            if i == 1:
+                rangeGsStr = startLetter + str(startNum - 1)
+            else:
+                rangeGsStr = startLetter + str(startNum_old) + ":" + startLetter + str(startNum - 1)
+
+        
+
+        # 计算填充范围
+        endLetter = jsonData_value_sourceInfo["range"].get("endLetter",startLetter)
+        endNum = jsonData_value_sourceInfo["range"].get("endNum",str(nrows))
+        rangeEndStr = endLetter + endNum
+
+        # 公式范围
+        print("rangeGsStr",rangeGsStr)
+        print("fillRangeStr",startLetter + str(startNum) + ":" + rangeEndStr)
+
+        sourceRange = sht.range(rangeGsStr).api
+        # 填充范围
+        fillRange = sht.range(startLetter + str(startNum_old) + ":" + rangeEndStr).api
+        
+        
 
-    # 新建Shell
-    def addShell(self,jsonData):
-        # print("新建Shell",jsonData)
         print_data = [
-            {"text": "新建Shell,请稍后...", "color": "33"},
-            {"text": "[%s]%s,{%s}" % (jsonData["sourceTitle"],jsonData["after"],jsonData["targetShell"]), "color": "33"}
+            {"text": "填充公式", "color": "33"},
+            # {"text": "操作表格： [%s]%s!%s%s:%s" % (jsonData["sourceTitle"],jsonData["sourceShell"],jsonData["startRangeCol"],jsonData["startRangeRow"],rangeEndStr ), "color": "34"},
+            {"text": "正在计算中，请稍后...", "color": "33"}
 
         ]
         self.print_log(print_data, "INFO")
-        sht =  self.wb[jsonData["sourceTitle"]].sheets.add(jsonData["targetShell"], after=jsonData["after"])
+        # 执行填充
+        sourceRange.AutoFill(fillRange, 0)
+
+
+    # 新建Shell
+    def addShell(self,jsonData):
+        jsonDataAfter = jsonData.get("after",None)
+        # print("新建Shell",jsonData)
+        print_data = [
+            {"text": "新建Shell,请稍后...", "color": "33"},
+            {"text": "[%s]%s,{%s}" % (jsonData["sourceTitle"],jsonDataAfter,jsonData["targetShell"]), "color": "33"}
+
+        ]
+        
+        self.print_log(print_data, "INFO")
+        sht =  self.wb[jsonData["sourceTitle"]].sheets.add(jsonData["targetShell"], after=jsonDataAfter)
+
     
     # 筛选数据，计算数量，列
     def countyCountyNum(self,jsonData):
@@ -229,6 +335,7 @@ class BzttExcl:  #工单处理
         shtSource.range(jsonData["sourceRange"]["start"],jsonData["sourceRange"]["end"]).api.RemoveDuplicates(int(jsonData["RemoveNum"]))
     # 初始化数据 （格式化数据）
     def InitJson(self,formatJson):
+        print("InitJsonInitJsonInitJson")
         formatJsonStr =json.dumps(formatJson,ensure_ascii=False)   #转换字符串
 
         for i in range(0,len(formatJson["formatVariable"])):
@@ -242,7 +349,8 @@ class BzttExcl:  #工单处理
                 title = in_value.get("title","请输入：")
                 inputIntStr = input(title)
                 formatJsonStr = formatJsonStr.replace(info["name"],inputIntStr)
-
+                # 输入的变量保存
+                self.excel_data[info["name"]] = inputIntStr
             #字符串转换
             elif info["type"] == "replace":
                 #print("json:",info["value"])
@@ -267,8 +375,9 @@ class BzttExcl:  #工单处理
                     "change_unit":in_value.get("changeUnit","%Y%m%d"), # 变化单位,
                 }
                 target_time_str = self._change_time(in_value)
-                formatJsonStr = formatJsonStr.replace( "\"" + info["name"] + "\"",target_time_str)
-                
+                self.excel_data[info["name"]] = target_time_str
+                formatJsonStr = formatJsonStr.replace( "" + info["name"] + "","" + target_time_str + "")
+
             elif info["type"] == "date2Rang":
                 # print("date2Rang:",info["value"])
                 #str转time
@@ -320,10 +429,23 @@ class BzttExcl:  #工单处理
             {"text": print_str_paste, "color": "33"}
         ]
         self.print_log(print_data, "INFO")
-
+    
+        sourceRangeType = copyJson.get("sourceRangeType","range")
         shtSource = self.wb[copyJson["sourceTitle"]].sheets(copyJson["sourceShell"])
         shtTarget = self.wb[copyJson["targetTitle"]].sheets(copyJson["targetShell"])
-        rangSource = copyJson["sourceRange"]["start"] + ":" + copyJson["sourceRange"]["end"]
+        if sourceRangeType == "range":
+            rangSource = copyJson["sourceRange"]["start"] + ":" + copyJson["sourceRange"]["end"]
+        elif sourceRangeType == "sheet":
+            # sht = self.wb[jsonData["sourceTitle"]].sheets(jsonData["sourceShell"])
+            info = shtSource.used_range
+            nrows = info.last_cell.row
+            ncolumns = self.ExcelColumn(info.last_cell.column)
+            
+            rangSource = "A1:" + ncolumns + str(nrows)
+            print("nrows",nrows,"ncolumns",ncolumns,"rangSource",rangSource)
+            pass
+        else:
+            raise Exception("未知范围类型")
         shtSource.range(rangSource).copy(destination=None)
 
         for pasteInfo in copyJson["targetRange"]:
@@ -392,10 +514,11 @@ class BzttExcl:  #工单处理
                 startTimeStr = str(info["value"])
                 SourceTimeStr = datetime.datetime.strptime(startTimeStr,info["valueFormat"])  + datetime.timedelta(days = int(info["num"]))
                 startTimeStr = SourceTimeStr.strftime( info["format"] )   #'>%Y-%m-%d %H:%M:%S'
+                print("startTimeStr",startTimeStr)
                 formatJsonStr = formatJsonStr.replace( "\"" + info["name"] + "\"",startTimeStr)
 
             elif info["type"] == "date2Rang":
-                #print("date2Rang:",info["value"])
+                print("date2Rang:",info["value"])
                 #str转time
                 SourceTimeStr = datetime.datetime.strptime(str(info["value"]["dateStr"]),info["value"]["dateFormat"])
                 year = SourceTimeStr.year
@@ -438,18 +561,21 @@ class BzttExcl:  #工单处理
             {"text": "正在复制文件，请稍后...", "color": "33"}
         ]
         self.print_log(print_data, "INFO")
+        run_path = iniFile.get("runPath",os.getcwd() ) + "\\"
 
         #检测文件是否存在
 
         #复制文件,并重命名
         for fileInfo in iniFile["info"]:
+            targetName = fileInfo.get("targetName",fileInfo["name"])
+            targetPathAll = run_path + fileInfo["targetPath"] + targetName  + "." + fileInfo["suffix"]
             print_data = [
                 {"text": fileInfo["sourcePath"] + fileInfo["name"] + "." + fileInfo["suffix"], "color": "33","end":" --> "},
-                {"text": fileInfo["targetPath"] + fileInfo["name"] + "_" + fileInfo["title"] + "." + fileInfo["suffix"], "color": "33"},
+                {"text": targetPathAll, "color": "33"},
             ]
             self.print_log(print_data, "INFO")
-
-            shutil.copyfile(fileInfo["sourcePath"] + fileInfo["name"] + "." + fileInfo["suffix"], fileInfo["targetPath"] + fileInfo["name"] + "_" + fileInfo["title"] + "." + fileInfo["suffix"])
+            
+            shutil.copyfile(run_path + fileInfo["sourcePath"] + fileInfo["name"] + "." + fileInfo["suffix"],targetPathAll )
 
         print_data = [
             {"text": "文件复制完毕", "color": "33"}
@@ -467,20 +593,21 @@ class BzttExcl:  #工单处理
         display_alerts = appConfig.get("displayAlerts",False)
 
         self.app = xw.App(visible=visibleFlag,add_book=add_book)
-
         self.app.display_alerts=display_alerts
-        
+
+        print("初始化完成")
         #打开文件
         for fileInfo in iniFile["info"]:
+            targetName = fileInfo.get("targetName",fileInfo["name"])
+            targetPathAll = run_path + fileInfo["targetPath"] + targetName  + "." + fileInfo["suffix"]
             print_data = [
                 {"text": fileInfo["title"], "color": "33","end":" - "},
-                {"text": fileInfo["targetPath"] + fileInfo["name"] + "_" + fileInfo["title"] + "." + fileInfo["suffix"], "color": "33"},
-
+                {"text": targetPathAll, "color": "33"},
             ]
             self.print_log(print_data, "INFO")
 
             #self.wb[fileInfo["title"]] =  xw.Book(fileInfo["path"] + fileInfo["name"] + "_" + fileInfo["title"] + "." + fileInfo["suffix"]) #创建一个新的Excel文件
-            self.wb[fileInfo["title"]] =  self.app.books.open(fileInfo["targetPath"] + fileInfo["name"] + "_" + fileInfo["title"] + "." + fileInfo["suffix"]) #创建一个新的Excel文件
+            self.wb[fileInfo["title"]] =  self.app.books.open(targetPathAll) #创建一个新的Excel文件
         
         
         print_data = [
@@ -548,7 +675,8 @@ class BzttExcl:  #工单处理
                 startTime = self._change_time(info_value[0])
                 endTime = self._change_time(info_value[1])
                 dfData = dfData[(dfData[range_name] >= startTime) & (dfData[range_name] <= endTime)]
-            elif info["type"] == "regex": #正则筛选   
+            elif info["type"] == "regex": #正则筛选 
+
                 if info["condition"] == "in": # 包含
                     Bool = dfData[range_name].str.contains(info_value)
                     dfData = dfData[Bool]
@@ -595,9 +723,13 @@ class BzttExcl:  #工单处理
             {"text": "正在记录明细... ", "color": "33"}
         ]
         self.print_log(print_data, "INFO")
+        detailsRange = inData.get("detailsRange",[])
         # 循环写入明细数据
-        for details_range_info in inData["detailsRange"]:
+        for details_range_info in detailsRange:
             df_data_all_columns = details_range_info.get("columns",df_data_all.columns)
+            print("df_data_all_columns::",df_data_all_columns)
+            if len(df_data_all_columns) == 0:
+                df_data_all_columns = df_data_all.columns
             self.wb[details_range_info["targetTitle"]].sheets(details_range_info["targetShell"]). \
                 range(details_range_info["targetRange"]["startLetter"] + str(details_range_info["targetRange"]["startNum"])) \
                     .value=df_data_all[df_data_all_columns].set_index(df_data_all_columns[0])
@@ -648,12 +780,73 @@ class BzttExcl:  #工单处理
             SourceTimeStr = datetime.datetime.strptime(startTimeStr,in_format)  + datetime.timedelta(days = change_num)
         else:
             raise Exception("未知变化类型")
+        print("out_format",out_format)
         startTimeStr = SourceTimeStr.strftime( out_format )   #'>%Y-%m-%d %H:%M:%S'
         return startTimeStr
+    def DeleteData(self,inData):
+        print("未完成")
+    def InputData(self,inData):
+        inputData = inData.get("data",[])
+        for info in inputData:
+            print("inputData",info)
+            info_type = info.get("type","")
+            if info_type == "continue":
+                while 1:
+                    flag = input("请输入 Y 继续，或输入 N 结束").upper()
+                    if flag == "Y":
+                        break
+                    elif flag == "N":
+                        raise Exception("人工终止")
+    def GeneratePic(self,inData):
+        print("GeneratePic:",inData)
+        inDataValue = inData["value"]
 
+        sourceRangeType = inDataValue["sourceInfo"]["range"].get("sourceRangeType","range")
+
+        shtSource = self.wb[inDataValue["sourceInfo"]["title"]].sheets(inDataValue["sourceInfo"]["sheet"])
+
+        if sourceRangeType == "range":
+            rangSource = inDataValue["sourceInfo"]["range"]["startLetter"] + str(inDataValue["sourceInfo"]["range"]["startNum"]) + ":" + inDataValue["sourceInfo"]["range"]["endLetter"] + str(inDataValue["sourceInfo"]["range"]["endNum"])
+        elif sourceRangeType == "sheet":
+            # sht = self.wb[jsonData["sourceTitle"]].sheets(jsonData["sourceShell"])
+            info = shtSource.used_range
+            nrows = info.last_cell.row
+            ncolumns = self.ExcelColumn(info.last_cell.column)
+            
+            rangSource = "A1:" + ncolumns + str(nrows)
+            print("nrows",nrows,"ncolumns",ncolumns,"rangSource",rangSource)
+            pass
+        else:
+            raise Exception("未知范围类型")
+        shtSourceApi = shtSource.range(rangSource)
+        print("rangSourcerangSourcerangSource",rangSource)
+        # shtSourceApi = shtSource.range(
+        #         (1, 1),  # 获取 第一行 第一列
+        #         (13, 13)  # 获取 第 nrow 行 第 ncol 列
+        #     )
+
+
+        shtSourceApi.api.CopyPicture()                   # 复制图片区域
+        shtSource.api.Paste()                       # 粘贴
+
+        img_name=inDataValue.get('img_name',"默认数据")
+        pic=shtSource.pictures[0]                   #当前图片
+        pic.api.Copy()                          #复制图片
+
+        time.sleep(1)
+        img = ImageGrab.grabclipboard()         # 获取剪贴板的图片数据
+        img.save(img_name + ".png")             #保存图片
+        pic.delete()                            #删除sheet上的图片
+
+        pass
+
+
+
+        
     # 整体执行操作
     def performOperate(self,operate_json):
         for info in operate_json:
+            print("info",info)
             if info["type"] == "advancedFilter": # 高级筛选
                 print_data = [
                     {"text": "执行高级筛选：" + info.get("title","无描述"), "color": "33"}
@@ -668,8 +861,25 @@ class BzttExcl:  #工单处理
                 self.ifsData(info)
             if info["type"] == "addShell":
                 self.addShell(info)
+            # 删除重复项
             if info["type"] == "RemoveDuplicates":
                 self.RemoveDuplicates(info)
+            # 填充公式
+            if info["type"] == "WriteGsData":
+                self.WriteGsData(info)
+            # 写入数组
+            if info["type"] == "WriteValue":
+                self.WriteValue(info)
+            # 删除 列或行
+            if info["type"] == "DeleteData":
+                self.DeleteData(info)
+            # 手动输入信息 Input
+            if info["type"] == "Input":
+                self.InputData(info)
+            # 生成图片
+            if info["type"] == "GeneratePic":
+                self.GeneratePic(info)
+
         pass
 
    
